// This file is an adaptation of
// <https://github.com/timothee-haudebourg/json-ld/blob/0.12.1/core/src/loader/reqwest/content_type.rs>.
//
// Copyright (C) 2022 Timothée Haudebourg et al.
// Licensed under either the Apache License, Version 2.0; or the MIT License.
// See <https://github.com/timothee-haudebourg/json-ld/tree/0.12.1> for details.
//
// Modified by anna <owo@fef.moe> to accept the `application/activity+json`
// MIME type as proper JSON-LD.

use hashbrown::HashMap;
use mime::Mime;
use reqwest::header::HeaderValue;
use std::str::FromStr;

/// Helper structure for parsing the `Content-Type` header for JSON-LD.
pub struct ContentType {
    mime: Mime,
    params: HashMap<Vec<u8>, Vec<u8>>,
}

impl ContentType {
    pub fn from_header(value: &HeaderValue) -> Option<ContentType> {
        enum State {
            Mime,
            NextParam,
            BeginKey,
            Key,
            BeginValue,
            QuotedValue,
            Value,
        }

        let mut state = State::Mime;
        let mut mime = Vec::new();
        let mut current_key = Vec::new();
        let mut current_value = Vec::new();
        let mut params = HashMap::new();

        let mut bytes = value.as_bytes().iter();

        loop {
            match state {
                State::Mime => match bytes.next().copied() {
                    Some(b';') => state = State::BeginKey,
                    Some(b) => mime.push(b),
                    None => break,
                },
                State::NextParam => match bytes.next().copied() {
                    Some(b';') => state = State::BeginKey,
                    Some(_) => return None,
                    None => break,
                },
                State::BeginKey => match bytes.next().copied() {
                    Some(b' ') => {}
                    Some(b) => {
                        current_key.push(b);
                        state = State::Key;
                    }
                    None => return None,
                },
                State::Key => match bytes.next().copied() {
                    Some(b'=') => state = State::BeginValue,
                    Some(b) => current_key.push(b),
                    None => return None,
                },
                State::BeginValue => match bytes.next().copied() {
                    Some(b'"') => state = State::QuotedValue,
                    Some(b) => {
                        state = State::Value;
                        current_value.push(b);
                    }
                    _ => return None,
                },
                State::QuotedValue => match bytes.next().copied() {
                    Some(b'"') => {
                        params.insert(
                            std::mem::take(&mut current_key),
                            std::mem::take(&mut current_value),
                        );
                        state = State::NextParam;
                    }
                    Some(b) => current_value.push(b),
                    None => return None,
                },
                State::Value => match bytes.next().copied() {
                    Some(b';') => {
                        params.insert(
                            std::mem::take(&mut current_key),
                            std::mem::take(&mut current_value),
                        );
                        state = State::BeginKey;
                    }
                    Some(b) => current_value.push(b),
                    None => {
                        params.insert(
                            std::mem::take(&mut current_key),
                            std::mem::take(&mut current_value),
                        );
                        break;
                    }
                },
            }
        }

        Mime::from_str(std::str::from_utf8(&mime).ok()?)
            .map(|mime| ContentType { mime, params })
            .ok()
    }

    pub fn is_json_ld(&self) -> bool {
        [
            "application/activity+json",
            "application/ld+json",
            "application/json",
        ]
        .iter()
        .any(|mime| *mime == self.mime)
    }

    pub fn is_proper_json_ld(&self) -> bool {
        ["application/activity+json", "application/ld+json"]
            .iter()
            .any(|mime| *mime == self.mime)
    }

    pub fn mime(&self) -> &Mime {
        &self.mime
    }

    pub fn into_mime(self) -> Mime {
        self.mime
    }

    pub fn profile(&self) -> Option<&[u8]> {
        self.params.get(b"profile".as_slice()).map(Vec::as_slice)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn parse_content_type_1() {
        let content_type = ContentType::from_header(
            &HeaderValue::from_str(
                "application/ld+json;profile=http://www.w3.org/ns/json-ld#expanded",
            )
            .unwrap(),
        )
        .unwrap();
        assert_eq!(*content_type.mime(), "application/ld+json");
        assert_eq!(
            content_type.profile(),
            Some(b"http://www.w3.org/ns/json-ld#expanded".as_slice())
        );
    }

    #[test]
    fn parse_content_type_2() {
        let content_type = ContentType::from_header(
            &HeaderValue::from_str(
                "application/ld+json; profile=http://www.w3.org/ns/json-ld#expanded",
            )
            .unwrap(),
        )
        .unwrap();
        assert_eq!(*content_type.mime(), "application/ld+json");
        assert_eq!(
            content_type.profile(),
            Some(b"http://www.w3.org/ns/json-ld#expanded".as_slice())
        );
    }

    #[test]
    fn parse_content_type_3() {
        let content_type = ContentType::from_header(
            &HeaderValue::from_str(
                "application/ld+json; profile=http://www.w3.org/ns/json-ld#expanded; q=1",
            )
            .unwrap(),
        )
        .unwrap();
        assert_eq!(*content_type.mime(), "application/ld+json");
        assert_eq!(
            content_type.profile(),
            Some(b"http://www.w3.org/ns/json-ld#expanded".as_slice())
        );
    }

    #[test]
    fn parse_content_type_4() {
        let content_type = ContentType::from_header(
            &HeaderValue::from_str(
                "application/ld+json; profile=\"http://www.w3.org/ns/json-ld#expanded\"; q=1",
            )
            .unwrap(),
        )
        .unwrap();
        assert_eq!(*content_type.mime(), "application/ld+json");
        assert_eq!(
            content_type.profile(),
            Some(b"http://www.w3.org/ns/json-ld#expanded".as_slice())
        );
    }

    #[test]
    fn parse_content_type_5() {
        let content_type = ContentType::from_header(
            &HeaderValue::from_str(
                "application/ld+json; profile=\"http://www.w3.org/ns/json-ld#expanded\"",
            )
            .unwrap(),
        )
        .unwrap();
        assert_eq!(*content_type.mime(), "application/ld+json");
        assert_eq!(
            content_type.profile(),
            Some(b"http://www.w3.org/ns/json-ld#expanded".as_slice())
        );
    }

    #[test]
    fn parse_content_type_6() {
        let content_type = ContentType::from_header(
            &HeaderValue::from_str(
                "application/ld+json;profile=\"http://www.w3.org/ns/json-ld#expanded\"; q=1",
            )
            .unwrap(),
        )
        .unwrap();
        assert_eq!(*content_type.mime(), "application/ld+json");
        assert_eq!(
            content_type.profile(),
            Some(b"http://www.w3.org/ns/json-ld#expanded".as_slice())
        );
    }

    #[test]
    fn parse_content_type_7() {
        let content_type = ContentType::from_header(&HeaderValue::from_str("application/ld+json; profile=\"http://www.w3.org/ns/json-ld#flattened http://www.w3.org/ns/json-ld#compacted\"; q=1").unwrap()).unwrap();
        assert_eq!(*content_type.mime(), "application/ld+json");
        assert_eq!(
            content_type.profile(),
            Some(
                b"http://www.w3.org/ns/json-ld#flattened http://www.w3.org/ns/json-ld#compacted"
                    .as_slice()
            )
        );
    }

    #[test]
    fn parse_content_type_8() {
        let content_type =
            ContentType::from_header(&HeaderValue::from_str("application/ld+json").unwrap())
                .unwrap();
        assert_eq!(*content_type.mime(), "application/ld+json");
    }

    #[test]
    fn is_proper_json_ld() {
        let content_type =
            ContentType::from_header(&HeaderValue::from_str("application/ld+json").unwrap())
                .unwrap();
        assert!(content_type.is_json_ld());
        assert!(content_type.is_proper_json_ld());

        let content_type =
            ContentType::from_header(&HeaderValue::from_str("application/activity+json").unwrap())
                .unwrap();
        assert!(content_type.is_json_ld());
        assert!(content_type.is_proper_json_ld());
    }

    #[test]
    fn is_json_ld() {
        let content_type =
            ContentType::from_header(&HeaderValue::from_str("application/json").unwrap()).unwrap();
        assert!(content_type.is_json_ld());
        assert!(!content_type.is_proper_json_ld());
    }

    #[test]
    fn is_not_json_ld() {
        let content_type =
            ContentType::from_header(&HeaderValue::from_str("application/xml").unwrap()).unwrap();
        assert!(!content_type.is_json_ld());
        assert!(!content_type.is_proper_json_ld());
    }
}
