// This file is an adaptation of
// <https://github.com/timothee-haudebourg/json-ld/blob/0.12.1/core/src/loader/reqwest/link.rs>.
//
// Copyright (C) 2022 Timothée Haudebourg et al.
// Licensed under either the Apache License, Version 2.0; or the MIT License.
// See <https://github.com/timothee-haudebourg/json-ld/tree/0.12.1> for details.

use iref::{IriRef, IriRefBuf};
use reqwest::header::HeaderValue;
use std::collections::HashMap;

pub struct Link {
    href: IriRefBuf,
    params: HashMap<Vec<u8>, Vec<u8>>,
}

impl Link {
    pub fn from_header(value: &HeaderValue) -> Option<Self> {
        enum State {
            BeginHref,
            Href,
            NextParam,
            BeginKey,
            Key,
            BeginValue,
            Value,
        }

        let mut state = State::BeginHref;
        let mut href = Vec::new();
        let mut current_key = Vec::new();
        let mut current_value = Vec::new();
        let mut params = HashMap::new();

        let mut bytes = value.as_bytes().iter();

        loop {
            match state {
                State::BeginHref => match bytes.next().copied() {
                    Some(b'<') => state = State::Href,
                    _ => break None,
                },
                State::Href => match bytes.next().copied() {
                    Some(b'>') => state = State::NextParam,
                    Some(b) => {
                        href.push(b);
                    }
                    None => break None,
                },
                State::NextParam => match bytes.next().copied() {
                    Some(b';') => state = State::BeginKey,
                    Some(_) => break None,
                    None => {
                        break match IriRefBuf::from_vec(href) {
                            Ok(href) => Some(Self { href, params }),
                            Err(_) => None,
                        }
                    }
                },
                State::BeginKey => match bytes.next().copied() {
                    Some(b' ') => {}
                    Some(b) => {
                        current_key.push(b);
                        state = State::Key
                    }
                    None => break None,
                },
                State::Key => match bytes.next().copied() {
                    Some(b'=') => state = State::BeginValue,
                    Some(b) => current_key.push(b),
                    None => break None,
                },
                State::BeginValue => match bytes.next().copied() {
                    Some(b'"') => state = State::Value,
                    _ => break None,
                },
                State::Value => match bytes.next().copied() {
                    Some(b'"') => {
                        params.insert(
                            std::mem::take(&mut current_key),
                            std::mem::take(&mut current_value),
                        );

                        state = State::NextParam
                    }
                    Some(b) => current_value.push(b),
                    None => break None,
                },
            }
        }
    }

    pub fn href(&self) -> IriRef {
        self.href.as_iri_ref()
    }

    pub fn rel(&self) -> Option<&[u8]> {
        self.params.get(b"rel".as_slice()).map(Vec::as_slice)
    }

    pub fn typ(&self) -> Option<&[u8]> {
        self.params.get(b"type".as_slice()).map(Vec::as_slice)
    }

    pub fn is_proper_json_ld(&self) -> bool {
        self.typ()
            .map(|typ| typ == b"application/activity+json" || typ == b"application/ld+json")
            .unwrap_or(false)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn parse_link_1() {
        let link = Link::from_header(
            &HeaderValue::from_str(
                "<http://www.example.org/context>; rel=\"context\"; type=\"application/ld+json\"",
            )
            .unwrap(),
        )
        .unwrap();
        assert_eq!(link.href(), "http://www.example.org/context");
        assert_eq!(link.rel(), Some(b"context".as_slice()));
        assert_eq!(link.typ(), Some(b"application/ld+json".as_slice()))
    }

    #[test]
    fn parse_link_2() {
        let link = Link::from_header(&HeaderValue::from_str("<http://www.example.org/context>; rel=\"context\"; type=\"application/ld+json\"; foo=\"bar\"").unwrap()).unwrap();
        assert_eq!(link.href(), "http://www.example.org/context");
        assert_eq!(link.rel(), Some(b"context".as_slice()));
        assert_eq!(link.typ(), Some(b"application/ld+json".as_slice()))
    }

    #[test]
    fn parse_link_3() {
        let link =
            Link::from_header(&HeaderValue::from_str("<http://www.example.org/context>").unwrap())
                .unwrap();
        assert_eq!(link.href(), "http://www.example.org/context")
    }

    #[test]
    fn is_proper_json_ld() {
        let link = Link::from_header(
            &HeaderValue::from_str(
                "<http://www.example.org/context>; rel=\"context\"; type=\"application/ld+json\"",
            )
            .unwrap(),
        )
        .unwrap();
        assert!(link.is_proper_json_ld());

        let link = Link::from_header(
            &HeaderValue::from_str(
                "<http://www.example.org/context>; rel=\"context\"; type=\"application/activity+json\"",
            )
                .unwrap(),
        )
            .unwrap();
        assert!(link.is_proper_json_ld());

        let link = Link::from_header(
            &HeaderValue::from_str(
                "<http://www.example.org/context>; rel=\"context\"; type=\"application/json\"",
            )
            .unwrap(),
        )
        .unwrap();
        assert!(!link.is_proper_json_ld());
    }
}
