use jsonwebtoken::{DecodingKey, EncodingKey};
use std::env;

/// Central repository for the backend's config.
pub struct Config {
    pub bind_addr: String,
    pub bind_port: u16,
    pub database_url: String,
    pub jwt_encode_key: EncodingKey,
    pub jwt_decode_key: DecodingKey,
    pub local_domain: String,
}

impl Config {
    pub fn from_env() -> Config {
        let jwt_secret = parse_sha256(
            env::var("JWT_SECRET")
                .expect("JWT_SECRET must be a 32-byte base-16 string")
                .as_str(),
        );
        Config {
            bind_addr: env::var("BIND_ADDR").unwrap_or("localhost".into()),
            bind_port: env::var("BIND_PORT")
                .unwrap_or("1312".into())
                .parse()
                .expect("BIND_PORT is invalid"),
            database_url: env::var("DATABASE_URL")
                .unwrap_or("postgres://nyanoblog@localhost/nyanoblog".into()),
            jwt_encode_key: EncodingKey::from_secret(jwt_secret.as_slice()),
            jwt_decode_key: DecodingKey::from_secret(jwt_secret.as_slice()),
            local_domain: env::var("LOCAL_DOMAIN").expect("LOCAL_DOMAIN not specified"),
        }
    }
}

fn parse_sha256(s: &str) -> Vec<u8> {
    if s.len() != 64 {
        panic!("SHA256 must be a 64-character hex string");
    }
    (0..64)
        .step_by(2)
        .map(|i| {
            u8::from_str_radix(&s[i..i + 2], 16).expect("SHA256 must be a 64-character hex string")
        })
        .collect()
}
