use sqlx::{
    database::{HasArguments, HasValueRef},
    encode::IsNull,
    error::BoxDynError,
    Database,
};
use std::{
    any::type_name,
    fmt,
    hash::{Hash, Hasher},
    marker::PhantomData,
};

/// Primitive type for regular database primary keys.
pub type IntId = i64;

/// Safe ID type for database models.
/// `T` is the database model that this ID references,
/// `I` is the primitive type representing the raw database type.
///
/// Database models should define their own ID types, like so:
///
/// ```
/// pub type MyId = Id<MyModel>;
///
/// pub struct MyModel {
///     pub id: MyId,
///     // ...
/// }
/// ```
pub struct Id<T: ?Sized, I = IntId> {
    id: I,
    _phantom: PhantomData<T>,
}

impl<T: ?Sized, I> Id<T, I> {
    pub fn inner(&self) -> &I {
        &self.id
    }

    pub fn into_inner(self) -> I {
        self.id
    }

    pub const fn from(id: I) -> Self {
        Id {
            id,
            _phantom: PhantomData,
        }
    }
}

impl<T: ?Sized, I> Clone for Id<T, I>
where
    I: Clone,
{
    fn clone(&self) -> Self {
        Id {
            id: self.id.clone(),
            _phantom: PhantomData,
        }
    }
}
impl<T: ?Sized, I> Copy for Id<T, I> where I: Copy {}

impl<T: ?Sized, I> From<I> for Id<T, I> {
    fn from(id: I) -> Self {
        Id {
            id,
            _phantom: PhantomData,
        }
    }
}

impl<T: ?Sized, I> PartialEq for Id<T, I>
where
    I: PartialEq,
{
    fn eq(&self, other: &Self) -> bool {
        self.id.eq(&other.id)
    }
}
impl<T: ?Sized, I> Eq for Id<T, I> where I: Eq {}

impl<T: ?Sized, I> Hash for Id<T, I>
where
    I: Hash,
{
    fn hash<H>(&self, state: &mut H)
    where
        H: Hasher,
    {
        self.id.hash(state);
    }
}

unsafe impl<T: ?Sized, I> Send for Id<T, I> where I: Send {}
unsafe impl<T: ?Sized, I> Sync for Id<T, I> where I: Sync {}

impl<'r, DB: Database, T: ?Sized, I> sqlx::Decode<'r, DB> for Id<T, I>
where
    I: sqlx::Decode<'r, DB>,
{
    fn decode(value: <DB as HasValueRef<'r>>::ValueRef) -> Result<Self, BoxDynError> {
        let id = I::decode(value)?;
        Ok(Id {
            id,
            _phantom: PhantomData,
        })
    }
}

impl<'q, DB: Database, T: ?Sized, I> sqlx::Encode<'q, DB> for Id<T, I>
where
    I: sqlx::Encode<'q, DB>,
{
    fn encode(self, buf: &mut <DB as HasArguments<'q>>::ArgumentBuffer) -> IsNull {
        self.id.encode(buf)
    }

    fn encode_by_ref(&self, buf: &mut <DB as HasArguments<'q>>::ArgumentBuffer) -> IsNull {
        self.id.encode_by_ref(buf)
    }

    fn produces(&self) -> Option<DB::TypeInfo> {
        self.id.produces()
    }

    fn size_hint(&self) -> usize {
        self.id.size_hint()
    }
}

impl<DB: Database, T: ?Sized, I> sqlx::Type<DB> for Id<T, I>
where
    I: sqlx::Type<DB>,
{
    fn type_info() -> DB::TypeInfo {
        I::type_info()
    }

    fn compatible(ty: &DB::TypeInfo) -> bool {
        I::compatible(ty)
    }
}

impl<T: ?Sized, I> serde::Serialize for Id<T, I>
where
    I: serde::Serialize,
{
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        self.id.serialize(serializer)
    }
}

impl<'de, T: ?Sized, I> serde::Deserialize<'de> for Id<T, I>
where
    I: serde::Deserialize<'de>,
{
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        let id = I::deserialize(deserializer)?;
        Ok(Id {
            id,
            _phantom: PhantomData,
        })
    }

    fn deserialize_in_place<D>(deserializer: D, place: &mut Self) -> Result<(), D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        I::deserialize_in_place(deserializer, &mut place.id)
    }
}

impl<T: ?Sized, I> fmt::Debug for Id<T, I>
where
    I: fmt::Debug,
{
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}#{:?}", type_name::<T>(), &self.id)
    }
}

impl<T: ?Sized, I> fmt::Display for Id<T, I>
where
    I: fmt::Display,
{
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        fmt::Display::fmt(&self.id, f)
    }
}
