use sqlx::PgPool;

use crate::core::*;
use crate::data::Count;
use crate::model::AccountId;
use crate::model::{Follow, FollowId, NewFollow};

pub struct PgFollowDataSource {
    pool: PgPool,
}

impl PgFollowDataSource {
    pub fn new(pool: PgPool) -> PgFollowDataSource {
        PgFollowDataSource { pool }
    }

    pub async fn by_id(&self, id: FollowId) -> Result<Follow> {
        let (follower_id, followee_id) = id.into_inner();
        self.by_follower_and_followee(follower_id, followee_id)
            .await
    }

    pub async fn create(&self, new: NewFollow) -> Result<Follow> {
        let follow: Follow = sqlx::query_as(
            "INSERT INTO follows (iri, follower_id, followee_id)
            VALUES ($1, $2, $3)
            RETURNING *",
        )
        .bind(new.iri)
        .bind(new.follower_id)
        .bind(new.followee_id)
        .fetch_one(&self.pool)
        .await?;
        Ok(follow)
    }

    pub async fn by_follower_and_followee(
        &self,
        follower_id: AccountId,
        followee_id: AccountId,
    ) -> Result<Follow> {
        let follow: Follow =
            sqlx::query_as("SELECT * FROM follows WHERE follower_id = $1 AND followee_id = $2")
                .bind(follower_id)
                .bind(followee_id)
                .fetch_one(&self.pool)
                .await?;
        Ok(follow)
    }

    pub async fn followees_of(&self, account_id: AccountId) -> Result<Vec<Follow>> {
        let followees: Vec<Follow> = sqlx::query_as("SELECT * FROM follows WHERE follower_id = $1")
            .bind(account_id)
            .fetch_all(&self.pool)
            .await?;
        Ok(followees)
    }

    pub async fn followers_of(&self, account_id: AccountId) -> Result<Vec<Follow>> {
        let followers: Vec<Follow> = sqlx::query_as("SELECT * FROM follows WHERE followee_id = $1")
            .bind(account_id)
            .fetch_all(&self.pool)
            .await?;
        Ok(followers)
    }

    pub async fn following_count_of(&self, account_id: AccountId) -> Result<u32> {
        let followee_count: Count =
            sqlx::query_as("SELECT COUNT(*) AS count FROM follows WHERE follower_id = $1")
                .bind(account_id)
                .fetch_one(&self.pool)
                .await?;
        Ok(followee_count.count as u32)
    }

    pub async fn follower_count_of(&self, account_id: AccountId) -> Result<u32> {
        let follower_count: Count =
            sqlx::query_as("SELECT COUNT(*) AS count FROM follows WHERE followee_id = $1")
                .bind(account_id)
                .fetch_one(&self.pool)
                .await?;
        Ok(follower_count.count as u32)
    }
}
