use crate::data::memcache::{Indexable, MemCache};
use crate::model::{AccountId, Like, LikeId, NoteId};

pub struct MemLikeDataSource {
    cache: MemCache<LikeId, Like>,
    counts: MemCache<NoteId, LikeCount>,
}

#[derive(Clone)]
struct LikeCount {
    note_id: NoteId,
    likes: u64,
}

impl MemLikeDataSource {
    pub fn new() -> MemLikeDataSource {
        MemLikeDataSource {
            cache: MemCache::new(),
            counts: MemCache::new(),
        }
    }

    pub async fn by_note_and_account(
        &self,
        note_id: NoteId,
        account_id: AccountId,
    ) -> Option<Like> {
        self.cache
            .get((note_id, account_id).into())
            .await
            .map(|arc| arc.as_ref().clone())
    }

    pub async fn store(&self, like: Like) {
        self.cache.put(like).await;
    }

    /// Delete a like entry from the cache **without decrementing the cached count**.
    pub async fn delete(&self, note_id: NoteId, account_id: AccountId) {
        self.cache.del((note_id, account_id).into()).await;
    }

    pub async fn get_count_by_note(&self, note_id: NoteId) -> Option<u64> {
        self.counts.get(note_id).await.map(|count| count.likes)
    }

    pub async fn store_count_by_note(&self, note_id: NoteId, count: u64) {
        self.counts
            .put(LikeCount {
                note_id,
                likes: count,
            })
            .await;
    }

    pub async fn increment_count(&self, note_id: NoteId) {
        self.counts.update(note_id, |entry| entry.likes += 1).await;
    }

    pub async fn decrement_count(&self, note_id: NoteId) {
        self.counts.update(note_id, |entry| entry.likes -= 1).await;
    }
}

impl Indexable<LikeId> for Like {
    fn get_id(&self) -> LikeId {
        (self.note_id, self.account_id).into()
    }
}

impl Indexable<NoteId> for LikeCount {
    fn get_id(&self) -> NoteId {
        self.note_id
    }
}
