use chrono::prelude::*;
use sqlx::PgPool;

use crate::core::*;
use crate::data::Count;
use crate::model::{AccountId, NewNote, Note, NoteId};

pub struct PgNoteDataSource {
    pool: PgPool,
}

impl PgNoteDataSource {
    pub fn new(pool: PgPool) -> PgNoteDataSource {
        PgNoteDataSource { pool }
    }

    pub async fn by_id(&self, id: NoteId) -> Result<Note> {
        let note: Note = sqlx::query_as("SELECT * FROM notes WHERE id = $1")
            .bind(id)
            .fetch_one(&self.pool)
            .await?;
        Ok(note)
    }

    pub async fn by_account(&self, id: AccountId, since: NaiveDateTime) -> Result<Vec<Note>> {
        let notes: Vec<Note> = sqlx::query_as(
            "SELECT * FROM notes
            WHERE account_id = $1 AND created_at < $1
            ORDER BY created_at
            LIMIT 20",
        )
        .bind(id)
        .bind(since)
        .fetch_all(&self.pool)
        .await?;
        Ok(notes)
    }

    pub async fn create(&self, new: NewNote) -> Result<Note> {
        let note: Note = sqlx::query_as(
            "INSERT INTO notes (
                account_id,
                iri,
                content,
                source,
                content_type,
                summary,
                sensitive
            ) VALUES ($1, $2, $3, $4, $5, $6, $7)
            RETURNING *",
        )
        .bind(new.account_id)
        .bind(new.iri)
        .bind(new.content)
        .bind(new.source)
        .bind(new.content_type)
        .bind(new.summary)
        .bind(new.sensitive)
        .fetch_one(&self.pool)
        .await?;
        Ok(note)
    }

    pub async fn delete(&self, id: NoteId) -> Result<()> {
        sqlx::query("DELETE FROM notes WHERE id = $1")
            .bind(id)
            .execute(&self.pool)
            .await?;
        Ok(())
    }

    pub async fn count_by_account(&self, account_id: AccountId) -> Result<u32> {
        let count: Count =
            sqlx::query_as("SELECT COUNT(*) AS count FROM notes WHERE account_id = $1")
                .bind(account_id)
                .fetch_one(&self.pool)
                .await?;
        Ok(count.count as u32)
    }
}
