use actix_rt::Arbiter;
use actix_web::web;
use async_trait::async_trait;
use serde::{Deserialize, Serialize};
use std::sync::atomic::{AtomicUsize, Ordering};
use std::sync::Weak;
use std::time::Duration;
use tokio::time::sleep;

use crate::core::*;
use crate::state::{AppState, State};

pub mod inbox;

const MAX_RETRIES: u32 = 3;
const ERROR_DELAY: Duration = Duration::from_secs(5 * 60);

/// A background job to be executed asynchronously.
///
/// Jobs run in a separate execution context, detached from the one that
/// enqueued it (which is usually a request handler).  This mechanism allows
/// for fast response times to requests and automatic retries on failures
/// such as network errors when a remote instance is currently offline.
///
/// [`Serialize`] and [`Deserialize`] are required for two reasons:
/// Firstly, the server may be restarted between waiting for the next retry of
/// failed jobs, in which case the parameters must be stored somewhere.
/// Secondly, a future version may support executing jobs on a remote node,
/// kind of like what Sidekiq does.
#[async_trait]
pub trait Job: Send + Sync + Serialize + Deserialize<'static> {
    /// Fulfill the errand.
    ///
    /// This will be called by the scheduler in a background thread.
    /// If the result is an error, the scheduler may retry it again later.
    async fn perform(&self, state: &AppState) -> Result<()>;

    /// Return this job's name.
    ///
    /// This should be unique per type of job, like for example the name of the
    /// struct implementing this trait.  Currently only used for logging.
    fn name(&self) -> &'static str;
}

pub struct Scheduler {
    arbiter: Arbiter,
    counter: AtomicUsize,
    state: Weak<State>,
}

impl Scheduler {
    pub fn new(state: Weak<State>) -> Scheduler {
        Scheduler {
            arbiter: Arbiter::new(),
            counter: AtomicUsize::new(1),
            state,
        }
    }

    pub fn schedule<T>(&self, job: T)
    where
        T: 'static + Job,
    {
        let id = self.counter.fetch_add(1, Ordering::Relaxed);
        let state: AppState = web::Data::from(
            self.state
                .upgrade()
                .expect("Zombie scheduler (this is supposed to be impossible)"),
        );
        if !self.arbiter.spawn(worker(job, id, state)) {
            panic!("Arbiter has died");
        }
    }
}

async fn worker<T>(job: T, id: usize, state: AppState)
where
    T: Job,
{
    let name = job.name();
    for n in 1..=MAX_RETRIES {
        debug!(target: "job", "Starting job {name}#{id}");

        if let Err(e) = job.perform(&state).await {
            warn!(target: "job", "Job {name}#{id} failed (try {n}/{MAX_RETRIES}): {e}");

            if n == MAX_RETRIES {
                error!(target: "job", "Giving up on job {name}#{id} (retry limit reached)");
            } else {
                sleep(ERROR_DELAY).await;
            }
        } else {
            debug!(target: "job", "Job {name}#{id} succeeded");
            break;
        }
    }
}
