use actix_web::dev::{Payload, Service, ServiceRequest, ServiceResponse, Transform};
use actix_web::{FromRequest, HttpMessage, HttpRequest};
use futures::future::LocalBoxFuture;
use futures::FutureExt;
use std::future::{ready, Ready};
use std::rc::Rc;
use std::task::{Context, Poll};

use crate::core::*;
use crate::model::Account;
use crate::state::AppState;
use crate::util::token;

/// Factory for [`AuthMiddleware`].
pub struct Auth {
    state: AppState,
}

impl Auth {
    pub fn new(state: AppState) -> Auth {
        Auth { state }
    }
}

impl<S, B> Transform<S, ServiceRequest> for Auth
where
    S: Service<ServiceRequest, Response = ServiceResponse<B>, Error = actix_web::Error> + 'static,
{
    type Response = ServiceResponse<B>;
    type Error = actix_web::Error;
    type Transform = AuthMiddleware<S>;
    type InitError = ();
    type Future = Ready<std::result::Result<Self::Transform, Self::InitError>>;

    fn new_transform(&self, service: S) -> Self::Future {
        ready(Ok(AuthMiddleware {
            service: Rc::new(service),
            state: self.state.clone(),
        }))
    }
}

pub struct AuthMiddleware<S> {
    service: Rc<S>,
    state: AppState,
}

impl<S, B> Service<ServiceRequest> for AuthMiddleware<S>
where
    S: Service<ServiceRequest, Response = ServiceResponse<B>, Error = actix_web::Error> + 'static,
{
    type Response = ServiceResponse<B>;
    type Error = actix_web::Error;
    type Future = LocalBoxFuture<'static, std::result::Result<Self::Response, Self::Error>>;

    fn poll_ready(&self, ctx: &mut Context<'_>) -> Poll<std::result::Result<(), Self::Error>> {
        self.service.poll_ready(ctx)
    }

    fn call(&self, req: ServiceRequest) -> Self::Future {
        let service = self.service.clone();
        let state = self.state.clone();

        async move {
            let account = if let Some(token) = req.headers().get("Authorization") {
                let token = extract_token(token.to_str().unwrap())?;
                debug!("token = \"{}\"", token);
                let account = token::validate(&state, token).await?;
                Some(account)
            } else {
                None
            };
            req.extensions_mut().insert(account);
            service.call(req).await
        }
        .boxed_local()
    }
}

pub struct AuthData(Option<Account>);

impl AuthData {
    pub fn maybe(&self) -> Option<&Account> {
        self.0.as_ref()
    }

    pub fn require(&self) -> Result<&Account> {
        self.maybe().ok_or(Error::BadCredentials)
    }
}

impl FromRequest for AuthData {
    type Error = Error;
    type Future = Ready<std::result::Result<Self, Self::Error>>;

    fn from_request(req: &HttpRequest, _payload: &mut Payload) -> Self::Future {
        let val: Option<Option<Account>> = req.extensions_mut().remove();
        ready(match val {
            Some(a) => Ok(AuthData(a)),
            None => Err(Error::BadCredentials),
        })
    }
}

fn extract_token(header: &str) -> Result<&str> {
    const PREFIX: &'static str = "Bearer ";

    if header.starts_with(PREFIX) {
        Ok(&header[PREFIX.len()..])
    } else {
        Err(Error::BadCredentials)
    }
}
