use sqlx::PgPool;

use crate::core::*;
use crate::data::PgFollowDataSource;
use crate::model::{AccountId, Follow, FollowId, NewFollow};

pub struct FollowRepo {
    db: PgFollowDataSource,
}

impl FollowRepo {
    pub fn new(db_pool: PgPool) -> FollowRepo {
        FollowRepo {
            db: PgFollowDataSource::new(db_pool),
        }
    }

    pub async fn by_id<I>(&self, id: I) -> Result<Follow>
    where
        I: Into<FollowId> + Send,
    {
        self.db.by_id(id.into()).await
    }

    pub async fn create<T>(&self, new: T) -> Result<Follow>
    where
        T: Into<NewFollow> + Send,
    {
        self.db.create(new.into()).await
    }

    pub async fn by_follower_and_followee<I, J>(
        &self,
        follower_id: I,
        followee_id: J,
    ) -> Result<Follow>
    where
        I: Into<AccountId> + Send,
        J: Into<AccountId> + Send,
    {
        let (follower_id, followee_id) = (follower_id.into(), followee_id.into());
        self.db
            .by_follower_and_followee(follower_id, followee_id)
            .await
    }

    pub async fn follows<I, J>(&self, follower_id: I, followee_id: J) -> Result<bool>
    where
        I: Into<AccountId> + Send,
        J: Into<AccountId> + Send,
    {
        match self
            .by_follower_and_followee(follower_id, followee_id)
            .await
        {
            Ok(_) => Ok(true),
            Err(Error::NotFound) => Ok(false),
            Err(e) => Err(e),
        }
    }

    pub async fn followees_of<I>(&self, account_id: I) -> Result<Vec<Follow>>
    where
        I: Into<AccountId> + Send,
    {
        self.db.followees_of(account_id.into()).await
    }

    pub async fn followers_of<I>(&self, account_id: I) -> Result<Vec<Follow>>
    where
        I: Into<AccountId> + Send,
    {
        self.db.followers_of(account_id.into()).await
    }

    pub async fn following_count_of<I>(&self, account_id: I) -> Result<u32>
    where
        I: Into<AccountId> + Send,
    {
        self.db.following_count_of(account_id.into()).await
    }

    pub async fn follower_count_of<I>(&self, account_id: I) -> Result<u32>
    where
        I: Into<AccountId> + Send,
    {
        self.db.follower_count_of(account_id.into()).await
    }
}
