use sqlx::PgPool;

use crate::core::*;
use crate::data::{MemLikeDataSource, PgLikeDataSource};
use crate::model::{AccountId, Like, NewLike, NoteId};

pub struct LikeRepo {
    db: PgLikeDataSource,
    mem: MemLikeDataSource,
}

impl LikeRepo {
    pub fn new(db_pool: PgPool) -> LikeRepo {
        LikeRepo {
            db: PgLikeDataSource::new(db_pool),
            mem: MemLikeDataSource::new(),
        }
    }

    pub async fn create<T, E>(&self, new: T) -> Result<Like>
    where
        T: TryInto<Sane<NewLike>, Error = E> + Send,
        E: Into<Error>,
    {
        let new = new.try_into().map_err(|e| e.into())?.inner();
        let like = self.db.create(new).await?;
        self.mem.store(like.clone()).await;
        self.mem.increment_count(like.note_id).await;
        Ok(like)
    }

    pub async fn delete<N, A>(&self, note_id: N, account_id: A) -> Result<()>
    where
        N: Into<NoteId>,
        A: Into<AccountId>,
    {
        let (note_id, account_id) = (note_id.into(), account_id.into());
        self.db.delete(note_id, account_id).await?;
        self.mem.delete(note_id, account_id).await;
        self.mem.decrement_count(note_id).await;
        Ok(())
    }

    pub async fn count_by_note<I>(&self, note_id: I) -> Result<u64>
    where
        I: Into<NoteId> + Send,
    {
        let note_id = note_id.into();
        if let Some(count) = self.mem.get_count_by_note(note_id).await {
            return Ok(count);
        }

        let count = self.db.count_by_note(note_id).await?;
        self.mem.store_count_by_note(note_id, count).await;
        Ok(count)
    }

    pub async fn by_note_and_account<N, A>(&self, note_id: N, account_id: A) -> Result<Like>
    where
        N: Into<NoteId>,
        A: Into<AccountId>,
    {
        let (note_id, account_id) = (note_id.into(), account_id.into());
        if let Some(like) = self.mem.by_note_and_account(note_id, account_id).await {
            return Ok(like);
        }

        let like = self.db.by_note_and_account(note_id, account_id).await?;
        self.mem.store(like.clone()).await;
        Ok(like)
    }
}
