use chrono::NaiveDateTime;
use sqlx::PgPool;

use crate::core::*;
use crate::data::{MemNoteDataSource, PgNoteDataSource};
use crate::model::{AccountId, NewNote, Note, NoteId};

pub struct NoteRepo {
    db: PgNoteDataSource,
    mem: MemNoteDataSource,
}

impl NoteRepo {
    pub fn new(db_pool: PgPool) -> NoteRepo {
        NoteRepo {
            db: PgNoteDataSource::new(db_pool),
            mem: MemNoteDataSource::new(),
        }
    }

    pub async fn by_id<I>(&self, id: I) -> Result<Note>
    where
        I: Into<NoteId> + Send,
    {
        let id = id.into();
        match self.mem.by_id(id).await {
            Some(note) => Ok(note),
            None => self.db.by_id(id).await,
        }
    }

    pub async fn by_account<I>(&self, account_id: I, since: NaiveDateTime) -> Result<Vec<Note>>
    where
        I: Into<AccountId> + Send,
    {
        self.db.by_account(account_id.into(), since).await
    }

    pub async fn create<T, E>(&self, new: T) -> Result<Note>
    where
        T: TryInto<Sane<NewNote>, Error = E> + Send,
        E: Into<Error>,
    {
        let new = new.try_into().map_err(|e| e.into())?.inner();
        let note = self.db.create(new).await?;
        self.mem.store(note.clone()).await;
        Ok(note)
    }

    pub async fn delete<I>(&self, note_id: I) -> Result<()>
    where
        I: Into<NoteId>,
    {
        let note_id = note_id.into();
        self.db.delete(note_id).await?;
        self.mem.delete(note_id).await;
        Ok(())
    }

    pub async fn count_by_account<I>(&self, account_id: I) -> Result<u32>
    where
        I: Into<AccountId> + Send,
    {
        self.db.count_by_account(account_id.into()).await
    }
}
