use sqlx::PgPool;

use crate::core::*;
use crate::data::{MemObjectDataSource, PgObjectDataSource};
use crate::model::{NewObject, Object, ObjectId};

pub struct ObjectRepo {
    db: PgObjectDataSource,
    mem: MemObjectDataSource,
}

impl ObjectRepo {
    pub fn new(db_pool: PgPool) -> ObjectRepo {
        ObjectRepo {
            db: PgObjectDataSource::new(db_pool),
            mem: MemObjectDataSource::new(),
        }
    }

    pub async fn create(&self, new: NewObject) -> Result<Object> {
        let obj = self.db.create(new).await?;
        self.mem.store(obj.clone()).await;
        Ok(obj)
    }

    pub async fn update<T, E>(&self, object: T) -> Result<Object>
    where
        T: TryInto<Sane<Object>, Error = E>,
        E: Into<Error>,
    {
        let mut object = object.try_into().map_err(|e| e.into())?.inner();
        object.updated_at = utc_now();
        self.db.update(&object).await?;
        Ok(object)
    }

    pub async fn by_id<U>(&self, id: U) -> Result<Object>
    where
        U: Into<ObjectId>,
    {
        let id = id.into();
        match self.mem.by_id(id).await {
            Some(obj) => Ok(obj),
            None => self.db.by_id(id).await,
        }
    }

    pub async fn by_iri(&self, iri: &str) -> Result<Object> {
        self.db.by_iri(iri).await
    }
}
