use sqlx::PgPool;

use crate::core::*;
use crate::data::{MemUserDataSource, PgUserDataSource};
use crate::model::{AccountId, NewUser, User, UserId};

pub struct UserRepo {
    db: PgUserDataSource,
    mem: MemUserDataSource,
}

impl UserRepo {
    pub fn new(db_pool: PgPool) -> UserRepo {
        UserRepo {
            db: PgUserDataSource::new(db_pool),
            mem: MemUserDataSource::new(),
        }
    }

    pub async fn by_id<I>(&self, id: I) -> Result<User>
    where
        I: Into<UserId> + Send,
    {
        let id = id.into();
        match self.mem.by_id(id).await {
            Some(user) => Ok(user),
            None => self.db.by_id(id).await,
        }
    }

    pub async fn by_account<I>(&self, account_id: I) -> Result<User>
    where
        I: Into<AccountId> + Send,
    {
        self.db.by_account(account_id.into()).await
    }

    pub async fn by_email<T>(&self, email: T) -> Result<User>
    where
        T: AsRef<str>,
    {
        self.db.by_email(email.as_ref()).await
    }

    pub async fn create<T, E>(&self, new: T) -> Result<User>
    where
        T: TryInto<Sane<NewUser>, Error = E> + Send,
        E: Into<Error>,
    {
        let new = new.try_into().map_err(|e| e.into())?.inner();
        let user = self.db.create(new).await?;
        self.mem.store(user.clone()).await;
        Ok(user)
    }
}
