use actix_web::{delete, get, post, web, HttpResponse};
use serde::Deserialize;

use crate::core::*;
use crate::ent;
use crate::middle::AuthData;
use crate::model::{NewLike, NewNote, NoteId};
use crate::state::AppState;

#[derive(Deserialize)]
struct NoteData {
    status: String,
    sensitive: Option<bool>,
    spoiler_text: Option<String>,
}

#[post("")]
async fn create_note(
    data: web::Json<NoteData>,
    state: AppState,
    auth: AuthData,
) -> Result<HttpResponse> {
    let account = auth.require()?;
    let data = data.into_inner();
    // TODO: HTML escape stuff, i guess
    let new = Insane::from(NewNote {
        account_id: account.id,
        iri: None,
        content: data.status,
        source: None,
        content_type: None,
        sensitive: data.sensitive.unwrap_or(false),
        summary: data.spoiler_text,
    });
    let note = state.repo.notes.create(new).await?;
    Ok(HttpResponse::Ok().json(ent::Status::from_model(&state, &note, &auth).await?))
}

#[get("/{id}")]
async fn get_by_id(
    path: web::Path<NoteId>,
    state: AppState,
    auth: AuthData,
) -> Result<HttpResponse> {
    let id = path.into_inner();
    let note = state.repo.notes.by_id(id).await?;
    Ok(HttpResponse::Ok().json(ent::Status::from_model(&state, &note, &auth).await?))
}

#[delete("/{id}")]
async fn del_by_id(
    path: web::Path<NoteId>,
    state: AppState,
    auth: AuthData,
) -> Result<HttpResponse> {
    let id = path.into_inner();
    let auth_account = auth.require()?;
    let note = state.repo.notes.by_id(id).await?;
    if note.account_id != auth_account.id {
        return Ok(HttpResponse::Unauthorized().finish());
    }
    let response = ent::Status::from_model(&state, &note, &auth).await?;
    state.repo.notes.delete(id).await?;
    Ok(HttpResponse::Ok().json(response))
}

#[post("/{id}/favourite")]
async fn add_favourite(
    path: web::Path<NoteId>,
    state: AppState,
    auth: AuthData,
) -> Result<HttpResponse> {
    let note_id = path.into_inner();
    let auth_account = auth.require()?;
    let note = state.repo.notes.by_id(note_id).await?;
    if state
        .repo
        .likes
        .by_note_and_account(note_id, auth_account.id)
        .await
        .is_err()
    {
        state
            .repo
            .likes
            .create(Insane::from(NewLike {
                iri: None,
                note_id: note.id,
                account_id: auth_account.id,
            }))
            .await?;
    }
    let response = ent::Status::from_model(&state, &note, &auth).await?;
    Ok(HttpResponse::Ok().json(response))
}

#[post("/{id}/unfavourite")]
async fn del_favourite(
    path: web::Path<NoteId>,
    state: AppState,
    auth: AuthData,
) -> Result<HttpResponse> {
    let note_id = path.into_inner();
    let auth_account = auth.require()?;
    let note = state.repo.notes.by_id(note_id).await?;
    if state
        .repo
        .likes
        .by_note_and_account(note_id, auth_account.id)
        .await
        .is_ok()
    {
        state.repo.likes.delete(note.id, auth_account.id).await?;
    }
    let response = ent::Status::from_model(&state, &note, &auth).await?;
    Ok(HttpResponse::Ok().json(response))
}

pub fn configure(cfg: &mut web::ServiceConfig) {
    cfg.service(create_note)
        .service(get_by_id)
        .service(del_by_id);
}
