use serde::{de, Deserialize, Deserializer, Serialize, Serializer};
use std::fmt;
use std::iter::Peekable;
use std::str::Chars;

use crate::core::*;
use crate::util::validate::{ResultBuilder, Validate};

const MAX_PARAM_LEN: usize = 16384;

/// Bearer Capabilities URI, see <https://docs.joinmastodon.org/spec/bearcaps/>
#[derive(PartialEq, Debug)]
pub struct Bearcap {
    pub url: String,
    pub token: String,
}

struct MaybeBearcap {
    url: Option<String>,
    token: Option<String>,
}

impl TryFrom<&str> for Bearcap {
    type Error = Error;

    fn try_from(v: &str) -> Result<Self> {
        const BEAR_PREFIX: &'static str = "bear:";
        if !v.starts_with(BEAR_PREFIX) {
            return Err(Error::BadBearcap);
        }
        let v = &v[BEAR_PREFIX.len()..];

        let mut params = MaybeBearcap {
            url: None,
            token: None,
        };
        let mut chars = v.chars().peekable();
        for _ in 0..2 {
            match chars.next() {
                Some('?' | '&') => parse_param(&mut params, &mut chars)?,
                None => break,
                _ => return Err(Error::BadBearcap),
            }
        }

        params.try_into()
    }
}

fn parse_param(params: &mut MaybeBearcap, chars: &mut Peekable<Chars>) -> Result<()> {
    match chars.next() {
        Some('t') => params.token = Some(parse_param_val(chars)?),
        Some('u') => params.url = Some(parse_param_val(chars)?),
        _ => return Err(Error::BadBearcap),
    }

    Ok(())
}

fn parse_param_val(chars: &mut Peekable<Chars>) -> Result<String> {
    if chars.next() != Some('=') {
        return Err(Error::BadBearcap);
    }
    let mut val = String::new();

    for _ in 0..MAX_PARAM_LEN {
        let c = chars.peek().map(|c| *c);
        match c {
            Some('&') | None => return Ok(val),
            Some(c) => {
                chars.next();
                val.push(c);
            }
        }
    }

    warn!(
        "Rejecting bearcap with param longer than {} chars",
        MAX_PARAM_LEN
    );
    Err(Error::BadBearcap)
}

impl TryFrom<MaybeBearcap> for Bearcap {
    type Error = Error;

    fn try_from(v: MaybeBearcap) -> Result<Self> {
        if let (Some(token), Some(url)) = (v.token, v.url) {
            if token.len() > 0 && url.len() > 0 {
                return Ok(Bearcap { token, url });
            }
        }

        Err(Error::BadBearcap)
    }
}

impl Serialize for Bearcap {
    fn serialize<S>(&self, serializer: S) -> std::result::Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let s = format!("bear:?t={}&u={}", self.token, self.url);
        serializer.serialize_str(s.as_str())
    }
}

impl<'de> Deserialize<'de> for Bearcap {
    fn deserialize<D>(deserializer: D) -> std::result::Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        deserializer.deserialize_str(BearcapVisitor)
    }
}

struct BearcapVisitor;

impl<'de> de::Visitor<'de> for BearcapVisitor {
    type Value = Bearcap;

    fn expecting(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.write_str("a valid bearcap URI")
    }

    fn visit_str<E>(self, value: &str) -> std::result::Result<Self::Value, E>
    where
        E: de::Error,
    {
        match Bearcap::try_from(value) {
            Ok(bc) => Ok(bc),
            Err(_) => Err(E::custom("Expected a valid bearcap URI")),
        }
    }
}

impl Validate for Bearcap {
    fn validate(&self, builder: ResultBuilder) -> ResultBuilder {
        builder.field(&self.url, "url", |f| {
            f.check("URL must start with https://", |v| {
                v.starts_with("https://")
            })
        })
    }
}

#[cfg(test)]
mod tests {
    use crate::util::bear::Bearcap;
    use serde_test::{assert_tokens, Token};

    #[test]
    fn parse_valid() {
        let bc = Bearcap::try_from("bear:?t=asdf&u=https://example.com/test")
            .expect("Doesn't parse valid bearcap");
        assert_eq!(bc.url, String::from("https://example.com/test"));
        assert_eq!(bc.token, String::from("asdf"));
    }

    #[test]
    fn serialize() {
        let bc = Bearcap {
            token: String::from("asdf"),
            url: String::from("https://example.com/test"),
        };

        assert_tokens(
            &bc,
            &[Token::Str("bear:?t=asdf&u=https://example.com/test")],
        );
    }

    #[test]
    fn reject_without_url() {
        let result = Bearcap::try_from("bear:?t=asdf");
        assert!(result.is_err());
    }

    #[test]
    fn reject_without_token() {
        let result = Bearcap::try_from("bear:?u=https://example.com/test");
        assert!(result.is_err());
    }

    #[test]
    fn reject_with_empty_token() {
        let result = Bearcap::try_from("bear:?t=&u=https://example.com/test");
        assert!(result.is_err());
    }

    #[test]
    fn reject_with_empty_url() {
        let result = Bearcap::try_from("bear:?t=asdf&u=");
        assert!(result.is_err());
    }
}
