use argon2::password_hash::rand_core::OsRng;
use argon2::password_hash::SaltString;
use argon2::{Argon2, PasswordHash, PasswordHasher, PasswordVerifier};
use serde::{Deserialize, Serialize};

use crate::core::*;
use crate::util::validate::{FieldResultBuilder, ValidateField};

#[derive(Serialize, Deserialize, Clone)]
pub struct ClearPassword(String);

#[derive(sqlx::Type, Serialize, Deserialize, Clone)]
#[sqlx(transparent)]
pub struct HashedPassword(String);

impl HashedPassword {
    pub fn verify(&self, clear: &ClearPassword) -> Result<()> {
        let clear = clear.0.as_str();
        let hash = PasswordHash::new(self.0.as_str()).expect("Corrupt password hash in database");
        let argon2 = Argon2::default();
        argon2
            .verify_password(clear.as_bytes(), &hash)
            .map_err(|_| Error::BadCredentials)
    }
}

impl ClearPassword {
    pub fn hash(self) -> HashedPassword {
        let salt = SaltString::generate(&mut OsRng);
        let argon2 = Argon2::default();
        let hash = argon2.hash_password(self.0.as_bytes(), &salt).unwrap();
        HashedPassword(hash.to_string())
    }

    pub fn len(&self) -> usize {
        self.0.len()
    }

    // these also match non ascii chars but whatever

    pub fn has_uppercase(&self) -> bool {
        self.0.chars().any(|c| c.is_uppercase())
    }

    pub fn has_lowercase(&self) -> bool {
        self.0.chars().any(|c| c.is_lowercase())
    }

    pub fn has_digit(&self) -> bool {
        self.0.chars().any(|c| c.is_numeric())
    }

    pub fn has_special(&self) -> bool {
        self.0.chars().any(|c| !c.is_ascii_alphanumeric())
    }
}

impl From<String> for ClearPassword {
    fn from(val: String) -> ClearPassword {
        ClearPassword(val)
    }
}

impl From<String> for HashedPassword {
    fn from(val: String) -> HashedPassword {
        HashedPassword(val)
    }
}

impl From<&str> for HashedPassword {
    fn from(val: &str) -> HashedPassword {
        HashedPassword(String::from(val))
    }
}

impl<'a> ValidateField<'a> for ClearPassword {
    fn validate_field(builder: FieldResultBuilder<'a, Self>) -> FieldResultBuilder<'a, Self> {
        builder
            .check("Must be at least 8 characters long", |v| v.len() >= 8)
            .check("Must contain an uppercase character", |v| v.has_uppercase())
            .check("Must contain a lowercase character", |v| v.has_lowercase())
            .check("Must contain a digit", |v| v.has_digit())
            .check("Must contain a special character", |v| v.has_special())
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn validate_hashes() {
        let clear = ClearPassword::from("asdf".to_owned());
        let hash = clear.clone().hash();
        assert!(hash.verify(&clear).is_ok());
        assert!(hash.verify(&clear_from("fdsa")).is_err());
        assert!(hash.verify(&clear_from("asdf\0")).is_err());
    }

    fn clear_from(s: &str) -> ClearPassword {
        ClearPassword::from(s.to_owned())
    }
}
