use jsonwebtoken::{decode, encode, Algorithm, Header, Validation};
use serde::{Deserialize, Serialize};

use crate::core::*;
use crate::model::Account;
use crate::state::AppState;

const VALID_SECS: i64 = 30 * 86400; // 30 days

#[derive(Debug, Serialize, Deserialize)]
struct Claims {
    /// Account ID, stored as a string because JSON can't be trusted
    sub: String,
    /// Expiry date in UNIX time (seconds)
    exp: i64,
}

pub fn issue(state: &AppState, account: &Account) -> Result<String> {
    let claims = Claims {
        sub: format!("{}", account.id),
        exp: unix_now() + 30 * 86400,
    };
    let header = Header {
        kid: Some("signing_key".to_owned()),
        alg: Algorithm::HS256,
        ..Default::default()
    };

    Ok(encode(&header, &claims, &state.config.jwt_encode_key)?)
}

pub async fn validate(state: &AppState, token: &str) -> Result<Account> {
    let validation = Validation::new(Algorithm::HS256);
    let claims: Claims = decode(token, &state.config.jwt_decode_key, &validation)?.claims;
    if unix_now() - claims.exp > VALID_SECS {
        // tokens that are valid for longer than VALID_SECS are sus
        return Err(Error::BadCredentials);
    }
    let account_id: IntId = claims.sub.parse().expect("We issued an invalid token??");
    state.repo.accounts.by_id(account_id).await
}
