import {LitElement, html, css} from "lit";
import {customElement, property, query, state} from "lit/decorators.js";

import withSharedStyles from "../theme";

@customElement("nyano-button")
export default class NyanoButton extends LitElement {
    @property({type: Boolean, attribute: true})
    public disabled = false;

    @state()
    private pressed = false;

    @query("#btn")
    private btn: HTMLButtonElement;

    protected render() {
        return html`
            <button
                id="btn"
                ?disabled=${this.disabled}
                class="${this.pressed ? "pressed" : ""}"
                @mousedown="${this.handleMouseDown}"
                @mouseup="${this.handleMouseLeave}"
                @mouseleave="${this.handleMouseLeave}"
            >
                <slot/>
            </button>
        `;
    }

    private handleMouseDown(e: MouseEvent) {
        this.updateTransformOrigin(e);
        this.btn.addEventListener("mousemove", this.updateTransformOrigin);
        this.pressed = true;
    }

    private handleMouseLeave() {
        this.pressed = false;
        this.btn.removeEventListener("mousemove", this.updateTransformOrigin);
    }

    private updateTransformOrigin = (e: MouseEvent) => {
        const rect = this.btn.getBoundingClientRect();
        const center = {
            x: rect.width / 2,
            y: rect.height / 2,
        };
        const origin = {
            x: center.x + (e.clientX - rect.x - center.x) / 2,
            y: center.y + (e.clientY - rect.y - center.y) / 2,
        };
        this.btn.style.transformOrigin = `${origin.x}px ${origin.y}px`;
    }

    static get styles() {
        return withSharedStyles(css`
            :host {
                margin: 8px;
            }

            button {
                font: bold 1rem var(--font-family);
                text-transform: uppercase;
                background-color: var(--accent-color);
                color: var(--accent-text-color);
                padding: var(--border-radius) calc(var(--border-radius) * 2);
                border-radius: var(--border-radius);
                box-shadow:
                    rgba(0, 0, 0, 0.5) 0 -2px 10px 2px inset,
                    rgba(0, 0, 0, 0.4) 0  0    4px 0   inset;
                border: none;
                transition-duration: 50ms;
                transition-timing-function: ease-in-out;
                transition-property: background-color, color, filter, transform;
                transform-origin: center;
            }

            button.pressed {
                transform: scale(96%);
            }

            button:not(:disabled):hover {
                cursor: pointer;
                background-color: var(--accent-highlight-color);
                color: var(--accent-highlight-text-color);
            }

            button:disabled {
                filter: grayscale(78%) brightness(80%);
            }

            :host(.outline) button {
                background-color: transparent;
                border: 2px solid var(--accent-color);
                padding: calc(var(--border-radius) - 2px) calc(var(--border-radius) * 2 - 2px);
                color: var(--accent-color);
                transition-property: color, background-color, border-color, transform;
                box-shadow: none;
            }

            :host(.outline) button:not(:disabled):hover {
                border-color: var(--accent-highlight-color);
                background-color: var(--accent-highlight-color);
                color: var(--accent-highlight-text-color);
            }
        `);
    }
}
