import {LitElement, html, css} from "lit";
import {customElement, query, state} from "lit/decorators.js";
import {msg, str} from "@lit/localize";

import { maxNoteChars } from "../config";
import withSharedStyles from "../theme";

import "../components/button";
import "../components/icon-button";
import "../components/input-field";
import InputField from "../components/input-field";

@customElement("compose-form")
export default class ComposeForm extends LitElement {
    @query("#summary")
    private summaryInput: InputField;

    @state()
    private summary = "";

    @query("#content")
    private contentInput: HTMLTextAreaElement;

    @state()
    private content = "";

    private get remainingChars() {
        return maxNoteChars - this.summary.length - this.content.length;
    }

    protected render() {
        return html`
            <input-field
                id="summary"
                placeholder="${msg("Content warning (optional)")}"
                maxlength="${maxNoteChars - this.content.length}"
                @input=${() => this.summary = this.summaryInput.value}
            ></input-field>

            <textarea
                id="content"
                placeholder="${msg("Say something nice")}"
                @input=${() => this.content = this.contentInput.value}
                maxlength="${maxNoteChars - this.summary.length}"
            ></textarea>

            <div id="bar">
                <div id="actions">
                    <icon-button icon="paperclip" label="${msg("Add attachment")}"></icon-button>
                    <icon-button icon="list-ul" label="${msg("Add poll")}"></icon-button>
                    <icon-button icon="emoji-laughing" label="${msg("Insert emogi")}"></icon-button>
                    <icon-button icon="globe-local" label="${msg("Note visibility")}"></icon-button>
                </div>

                <div
                    id="remaining-chars"
                    class="${this.getRemainingCharsClass()}"
                    aria-description="${msg(str`${this.remainingChars} characters remaining`)}"
                >
                    ${this.remainingChars}
                </div>
                <nyano-button ?disabled=${this.content.length === 0}>${msg("Meow!")}</nyano-button>
            </div>
        `;
    }

    private getRemainingCharsClass() {
        const remaining = maxNoteChars - this.summary.length - this.content.length;
        if (remaining > 20) {
            return "";
        } else if (remaining > 0) {
            return "warning";
        } else {
            return "error";
        }
    }

    static get styles() {
        return withSharedStyles(css`
            textarea {
                display: block;
                width: calc(100% - 16px);
                height: 115px;
                padding: 8px;
                font-size: 1rem;
                font-family: var(--font-family);
                background-color: var(--background-lighter-color);
                color: var(--text-color);
                border: none;
                box-shadow: inset 0 1px 1px 0 rgba(255, 255, 255, 0.15);
                border-radius: var(--border-radius);
                resize: none;
                transition: box-shadow ease-in-out var(--anim-duration-short);
            }

            textarea:focus-visible {
                outline: 0;
                box-shadow:
                    inset 0 0 2px 1px var(--accent-highlight-color),
                          0 0 4px 0   var(--accent-highlight-color);
            }

            #bar {
                display: flex;
                flex-direction: row;
                padding-top: 16px;
            }

            #actions {
                flex: 1;
                display: flex;
                flex-direction: row;
                padding: 8px 0 0;
            }

            #actions > * {
                flex: 1;
            }

            #remaining-chars {
                padding: 8px 16px 8px 0;
                font-weight: bold;
            }

            #remaining-chars.almost-full {
                color: var(--warning-color);
            }

            #remaining-chars.full {
                color: var(--error-color);
            }

            nyano-button {
                margin: 0;
            }
        `);
    }
}
